/*********************/
/* Simple Deker test */
/*********************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdatomic.h>
#include <pthread.h>
#include "basic.h"

#define SZ (256)
static int verbose = 0 ;

typedef struct {
  int sz ;
  volatile int x ;
  volatile int want[2] ;
  volatile int turn ;
  volatile int start ;
} common_t ;

inline static void mbar(void) { __sync_synchronize(); }

inline static int other(int id) { return 1-id; }

inline static void lock(int id, common_t *q) {
  q->want[id] = 1 ;
  while (q->want[other(id)]) {
    if (q->turn != id) {
      q->want[id] = 0 ;
    }
    while (q->turn != id) ;
    q->want[id] = 1 ;
  }
}

inline static void unlock(int id, common_t *q) {
  q->turn = other(id);
  q->want[id] = 0 ;
}

typedef struct {
  int id ;
  common_t *common ;
} ctx_t ;

void *P(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  int id = p->id ;
  int sz = q->sz ;

  while (!q->start) ;

  for (int k = sz ; k > 0 ; k--) {
    lock(id,q) ;
    int tmp = q->x ;
    q->x = tmp+1 ;
    unlock(id,q) ;
  }
  return(NULL);
}


/*************************/
/* Thread lauch and join */
/*************************/

void run(int sz) {
  common_t c ;
  do {
    c.sz = sz ;
    c.want[0] = c.want[1] = 0 ;
    c.turn = 0 ;
    c.x = 0 ;
    c.start = 0 ;

    pthread_t th[2] ; ctx_t a[2] ;
    for (int k = 0 ; k < 2 ; k++) {
      ctx_t *p = &a[k] ;
      p->id = k ; p->common = &c ;
      create_thread(&th[k],P,p) ;
    }
    c.start = 1 ;
    for (int k = 0 ; k < 2 ; k++) join_thread(&th[k]) ;
    printf("x=%i\n",c.x) ;
  } while  (c.x == 2 * c.sz) ;
}


void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [n] [m]\n",pgm) ;
}


int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int sz = SZ ;


  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    sz = atoi(argv[0]) ;
  }
  run(sz) ;
  return 0 ;
}

